/* wptHTTP.cpp - Generic HTTP support
 *      Copyright (C) 2004-2007 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <ctype.h>
#include <errno.h>

#include "wptHTTP.h"
#include "wptTypes.h"
#include "wptErrors.h"
#include "StringBuffer.h"


/* Empty constructur to allow advanced requests. */
NetHTTP::NetHTTP (void)
{
    reset ();
}


/* Constructur to allow alternative syntax. */
NetHTTP::NetHTTP (const char *_host, int _port, const char *_url)
{
    reset ();
    this->host = strdup (_host);
    this->port = _port;
    this->url = strdup (_url);
    if (!sendRequest (_host, _port, _url))
	parseResponse (&statcode);

}


/* Standard constructur. */
NetHTTP::NetHTTP (const char *_url)
{
    reset ();
    open (_url);
}

/* Open a connection to the given URL @_url. */
int
NetHTTP::open (const char *_url)
{
    int err;

    /* A connection were privously established. */
    if (this->fd != 0)
	return 0;

    safe_free (this->host);
    safe_free (this->url);

    err = extractHostInfo (_url, &this->host, &this->url);
    if (err)
	return err;
    err = sendRequest (this->host, this->port, this->url);
    if (err)
	return err;
    return parseResponse (&statcode);
}


/* Reset object contents for first use. */
void
NetHTTP::reset (void)
{
    ver = 1;
    port = 80;
    host = url = NULL;
    req_headers = resp_headers = NULL;
    fd = statcode = 0;
    method = HTTP_GET;
    nleft = -1;
    error = 0;
}


/* Perform a HTTP 'HEAD' request. */
int
NetHTTP::head (const char *_url)
{
    int err;

    safe_free (host);
    safe_free (url);

    err = extractHostInfo (_url, &this->host, &this->url);
    if (err)
	return err;

    method = HTTP_HEAD;
    if (!sendRequest (host, port, url))
	err = parseResponse (&statcode);

    return err;
}


/* Perform a HTTP 'GET' request. */
int
NetHTTP::get (const char *_url)
{
    int err;

    safe_free (host);
    safe_free (url);

    err = extractHostInfo (_url, &this->host, &this->url);
    if (err)
	return err;

    method = HTTP_GET;
    if (!sendRequest (this->host, this->port, this->url))
	err = parseResponse (&statcode);

    return err;
}


/* Return HTTP status code. */
int 
NetHTTP::getStatusCode (void)
{
    return statcode;
}


/* Return MIME content type. */
const char*
NetHTTP::getContentType (void)
{
    const char *type = NULL;

    findHeader (resp_headers, "Content-Type", &type);
    return type;
}


/* Return content length. */
unsigned int
NetHTTP::getContentLength (void)
{
    const char *len = NULL;

    if (findHeader (resp_headers, "Content-Length", &len))
	return strtoul (len, NULL, 10);
    return 0;
}


void
NetHTTP::addHeader (http_head_t *root, const char *val)
{
    http_head_t n, t;
    
    t = (http_head_t)calloc (1, sizeof *t+strlen (val)+2);
    if (!t)
	BUG (0);
    strcpy (t->d, val);

    if (!*root)
	*root = t;
    else {
	for (n = *root; n->next; n=n->next)
	    ;
	n->next = t;
    }
}


bool
NetHTTP::findHeader (http_head_t root, 
		     const char *name, const char **val)
{
    http_head_t n;
    char *p;
    
    *val = NULL;
    for (n = root; n; n = n->next) {
	if (strlen (n->d) >= strlen (name) &&
	    strstr (n->d, name)) {
	    p = strchr (n->d, ':');
            *val = p? n->d + (p - n->d + 1 + 1) : NULL;
	    return true;
        }
    }
    return false;
}


int
NetHTTP::connect (const char *_host, int _port)
{
    struct hostent *hp;
    struct sockaddr_in srv;
    unsigned long addr = 0;
    int i = 1;

    memset (&srv, 0, sizeof srv);
    srv.sin_port = htons ((unsigned short)_port);
    srv.sin_family = AF_INET;
    
    if (isalpha (*_host)) {
	hp = gethostbyname (_host);
	if (!hp) {
	    this->error = (int)WSAGetLastError ();
	    return WPTERR_WINSOCK_CONNECT;
	}
	memcpy (&srv.sin_addr, hp->h_addr, hp->h_length);
    }
    else {
	addr = inet_addr (_host);
	if (addr == INADDR_NONE) {
	    this->error = (int)WSAGetLastError ();
	    return WPTERR_WINSOCK_CONNECT;
	}
	memcpy (&srv.sin_addr, &addr, sizeof addr);
    }
	    
    fd = socket (AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_SOCKET;
    }

    if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, 
		    (const char*)&i, sizeof i)) {
	closesocket (fd);
	fd = 0;
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_SOCKET;
    }
    
    if (::connect (fd, (struct sockaddr *)&srv, sizeof (srv))) {
	closesocket (fd);
	fd = 0;
	this->error = (int)WSAGetLastError ();
	return WPTERR_WINSOCK_CONNECT;
    }

    return 0;
}


/* Return 0 if the next recv() call would be block. */
int
NetHTTP::isDataAvailable (int _fd)
{
    struct timeval tv;
    fd_set inp;
    int n;

    FD_ZERO (&inp);
    FD_SET (_fd, &inp);
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    n = select (_fd+1, &inp, NULL, NULL, &tv);
    if (n && FD_ISSET (_fd, &inp))
        return n;
    return 0;
}


/* Read a line from the open socket. A line means
   a sequence of octets terminated with a \r\n.
   Return value: 0 on success. */
int
NetHTTP::readLine (char *buf, unsigned int nbuf,
                 int nonblock, int *nn, int *eof)
{
    char c;
    int n, i;
    
    if (nn)
	*nn = 0;
    if (eof)
        *eof = 0;
    i = 0;
    do {
        if (nonblock == 1 && isDataAvailable (fd) == 0) {
            buf[i++] = '\0';
            i = -1;
            break;
        }
	n = recv (fd, &c, 1, 0);
	if (n == -1) {
	    this->error = (int)WSAGetLastError ();
	    break;
	}
	if (n == 0 || nbuf == 0 || c == '\n') {
	    if (n == 0) {
                if (eof)
                    *eof = 1;
                buf[i++] = '\0';
	    }
            else {
                buf[i++] = c;
                buf[i] = '\0';
            }
            break;
	}
        else {
            buf[i++] = c;
            nbuf--;
        }
    } while (n > 0 && i != -1);
    
    if (nn)
	*nn = i;
    return 0;
}



/* Extract the host from the given url @url. Return the host in
   @host and also return the resource part of the url in @new_url. */
int
NetHTTP::extractHostInfo (const char *_url, char **_host, char **new_url)
{   
    char tmpbuf[2*MAX_PATH+1];
    char *p;

    *_host = NULL;
    *new_url = NULL;

    /* XXX: do not use static buffers. */
    memset (tmpbuf, 0, sizeof (tmpbuf));
    strncpy (tmpbuf, _url, DIM (tmpbuf)-1);
    
    p = "http://";
    if (strlen (_url) < 10 || strncmp (_url, p, strlen (p)))
        return WPTERR_GENERAL;
    _url += strlen (p);
    p = strtok (tmpbuf+7, "/");
    if (!p)
        return WPTERR_GENERAL;
    *_host = strdup (p);   
    p = strchr (_url, '/');
    /* if no document were given so we assume the root '/'. */
    *new_url = !p? strdup ("/") : strdup (p);   

    return 0;
}


/* Add a request header. */
int 
NetHTTP::addRequestHeader (const char *name, const char *val)
{
    StringBuffer p;

    p = p + name + ": " + val;
    addHeader (&req_headers, p.getBuffer ());
    return 0;
}


void 
NetHTTP::setVersion (int _ver)
{
    if (_ver > 9)
	_ver = 0;
    this->ver = _ver;
}


int
NetHTTP::addRequestHeader (const char *name, unsigned int val)
{
    char buf[32];

    sprintf (buf, "%lu", (DWORD)val);
    return addRequestHeader (name, buf);
}


/* Prepare the request resource @url and send 
   it to host @host (port @port). */
int
NetHTTP::sendRequest (const char *_host, int _port, const char *_url)
{
    const char *id[] = {"GET", "HEAD", "PUT", "POST"};
    http_head_t h;
    StringBuffer out;
    int n;

    if (this->fd == 0) {
	int rc = connect (_host, _port);
	if (rc)
	    return rc;
    }

    if (*_url == '/')
    	_url++;
    if (_url == NULL)
    	_url = "";

    addRequestHeader ("Host", _host);
    if (ver < 1)
	addRequestHeader ("Connection", "close");

    out = out + id[method] + " " + "/" + _url + " " + "HTTP/1." + (int)ver + "\r\n";
    for (h = req_headers; h; h = h->next)
	out = out + h->d + "\r\n";
    out = out + "\r\n";

    n = write (out.getBuffer (), strlen (out.getBuffer ()));
    log_box ("debug", 0, "%d", n);
    return n > 0? 0 : WPTERR_GENERAL;
}


/* Parse all response resp_headers. */
int
NetHTTP::parseHeaders (http_head_t *r_head)
{
    char buf[300];
    int nn;
    int rc;
    
    if (!r_head)
	return WPTERR_GENERAL;

    do  {
	rc = readLine (buf, DIM (buf)-1, 1, &nn, NULL);
	if (rc)
	    return rc;
	if (nn == 2)
	    break; /* reach empty line */
	addHeader (r_head, buf);
    } while (rc == 0);
    return 0;
}


/* Default buffer size */
#define BUFSIZE 1024


/* Read data from the response and write it to @out. */
int
NetHTTP::readData (FILE *out)
{
    const char *val;
    char buf[BUFSIZE+1];
    int nlen, nn = 0, n, eof=0;
    int rc = 0;
    
    if (this->fd == 0 || this->resp_headers == NULL)
	return -1;

    nlen = getContentLength ();
    if (nlen == 0) {
        if (!findHeader (resp_headers, "Connection", &val) ||
	    strnicmp (val, "close", 4))
            return WPTERR_GENERAL;
    }

    val = getContentType ();
    if (!val)
        return WPTERR_GENERAL;
    if (strnicmp (val, "text", 4)) { /* binary */
	do {
	    n = recv (fd, buf, nlen > BUFSIZE? BUFSIZE : nlen, 0);
	    if (n == -1) {
		this->error = (int)WSAGetLastError ();
		return SOCKET_ERROR;
	    }
	    if (n == 0)
		break;
	    nlen -= n;
	    fwrite (buf, 1, n, out);
	} while (nlen > 0);
	return 0;
    }

    do {
	rc = readLine (buf, DIM (buf)-1, 1, &n, &eof);
	if (rc)
	    return rc;
	if (n > 0)
	    fwrite (buf, 1, n, out);
        if (nlen > 0) {
            nn += n;
            if (nlen == nn)
                break;
        }
    } while (eof == 0);

    return 0;
}


/* Parse the HTTP response line. */
int
NetHTTP::parseResponse (int *_statcode)
{
    http_head_t n;
    const char *tmp, *p;
    char buf[300];
    int code = 0, nn = 0;
    int rc;

    *_statcode = 0;
    rc = readLine (buf, DIM (buf)-1, 1, &nn, NULL);
    if (rc)
	return rc;

    tmp = "HTTP/1.";
    if (strlen (buf) < 8 || strncmp (buf, tmp, strlen (tmp)))
	return WPTERR_GENERAL;
    p = buf;
    p += strlen (tmp)+1 + 1; /* skip HTTP/1.x and WS */
    *_statcode = atoi (p);
    if (tmp == 0)
	return WPTERR_GENERAL;
    p += 3 + 1; /* number + WS */

    if (code == HTTP_STAT_400 ||            
	code == HTTP_STAT_403 ||
	code == HTTP_STAT_404 ||
	code == HTTP_STAT_405)
	return WPTERR_GENERAL;

    while (resp_headers) {
	n = resp_headers->next;
	free (resp_headers);
	resp_headers = n;
    }
    resp_headers = NULL;
    rc = parseHeaders (&resp_headers);
    if (rc)
	return rc;
    return 0;
}


/* Destroy HTTP object. */
NetHTTP::~NetHTTP (void)
{
    http_head_t h;
    
    while (resp_headers) {
	h = resp_headers->next;
	free (resp_headers);
	resp_headers = h;
    }
    while (req_headers) {
	h = req_headers->next;
	free (req_headers);
	req_headers = h;
    }
    safe_free (url);
    safe_free (host);
    if (fd != 0) {
	closesocket (fd);
	fd = 0;
    }
}


/* Read @buflen bytes from the http stream into the buffer @buf. */
int
NetHTTP::read (void *buf, unsigned int buflen)
{
    int n;

    if (this->fd == 0 || this->resp_headers == NULL)
	return -1;
    if (method != HTTP_GET)
	return -1;

    if (nleft == -1)
	nleft = getContentLength ();
    if ((int)buflen > nleft)
	buflen = nleft;
    if (nleft == 0)
	return -1;

    n = recv (fd, (char*)buf, (int)buflen, 0);
    if (n > 0) {
	nleft -= n;
	if (nleft < 0) 
	    nleft = 0;
    }
    return n;
}


/* Write the buffer @buf to the http stream.
   The function should be considered general as a replacement
   for send() to make sure the entire buffer is written. */
int
NetHTTP::write (const void *buf, unsigned int buflen)
{
    const char *buffer = (const char*)buf;
    unsigned int bytesleft = buflen;
    int nwritten;

    if (this->fd == 0)
	return -1;

    while (bytesleft > 0) {
	nwritten = send (this->fd, buffer, bytesleft, 0);
	if (nwritten == -1)
	    return -1;
	bytesleft -= nwritten;
	buffer += nwritten;
    }
    return buflen;
}


/* Return the Winsock specific error code. */
int NetHTTP::getErrorCode (void)
{
    return error;
}
