/* wptSOCKS.cpp - SOCKS proxy support
 *	Copyright (C) 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <windows.h>

#include "wptTypes.h"
#include "wptGPG.h"
#include "wptKeyserver.h"
#include "wptErrors.h"
#include "wptNLS.h"

/* Known socks errors. */
enum socks5_error_t {
    ERR_SUCCESS	    = 0x00,
    ERR_GENERAL	    = 0x01,
    ERR_NOTALLOWED  = 0x02,
    ERR_NETUNREACH  = 0x03,
    ERR_HOSTUNREACH = 0x04,
    ERR_CONNREFUSED = 0x05,
    ERR_INVADDR	    = 0x08
};

/* Known socks address types. */
enum socks5_addr_t {
    ADDR_IPV4	    = 0x01,
    ADDR_DOMAIN	    = 0x03,
    ADDR_IPV6	    = 0x04
};

/* Known socks authentication methods. */
enum socks5_method_t {
    METH_NOAUTH	    = 0x00,
    METH_GSSAPI	    = 0x01,
    METH_USERPASS   = 0x02,
    METH_ERR	    = 0xFF
};


/* Return human readable SOCKS error. */
static const char*
socks_strerror (int err)
{
    switch (err) {
    case ERR_SUCCESS: _("No error"); break;
    case ERR_GENERAL: _("General SOCKS error"); break;
    case ERR_NOTALLOWED: _("Connection not allowed by ruleset"); break;
    case ERR_NETUNREACH: _("Network is unreachable"); break;
    case ERR_HOSTUNREACH: _("Host could not be resolved"); break;
    case ERR_CONNREFUSED: _("Connection refused"); break;
    case ERR_INVADDR: _("Address type is not supported"); break;
    }
    return _("Unknown SOCKS error");
}


/* Send initial method packet. */
static int
send_methods (int conn_fd)
{
    BYTE pkt[4];

    pkt[0] = 0x05;
    pkt[1] = 0x02;
    pkt[2] = METH_NOAUTH;
    pkt[3] = METH_USERPASS;
    if (send (conn_fd, (const char*)pkt, DIM (pkt), 0) == SOCKET_ERROR) {
	log_debug ("send_methods failed ec=%d\n", (int)WSAGetLastError ());
	return -1;
    }
    return 0;
}


/* Parse the chosen server method. */
static int
recv_method (int conn_fd)
{
    BYTE pkt[2];

    if (recv (conn_fd, (char*)pkt, DIM (pkt), 0) == SOCKET_ERROR) {
	log_debug ("recv_method failed ec=%d\n", (int)WSAGetLastError ());
	return -1;
    }
    if (pkt[0] != 0x05) {
	log_debug ("recv_method: protocol violation\n");
	return -1;
    }
    return pkt[1];
}


/*
           +----+------+----------+------+----------+
           |VER | ULEN |  UNAME   | PLEN |  PASSWD  |
           +----+------+----------+------+----------+
           | 1  |  1   | 1 to 255 |  1   | 1 to 255 |
           +----+------+----------+------+----------+
    Do user/password authentication. 
*/
static int
do_userpass_auth (int conn_fd, keyserver_proxy_t ctx, int *err)
{
    BYTE *pkt;
    DWORD pktlen;

    *err = 0;
    pktlen = 1+1+strlen (ctx->user)+1+strlen (ctx->pass);
    pkt = (BYTE*)malloc (pktlen);
    pkt[0] = 0x05;
    pkt[1] = strlen (ctx->user);
    memcpy (pkt+2, ctx->user, strlen (ctx->user));
    pkt[2+strlen (ctx->user)] = strlen (ctx->pass);
    memcpy (pkt+2+strlen (ctx->user)+1, ctx->pass, strlen (ctx->pass));
    if (send (conn_fd, (const char*)pkt, pktlen, 0) == SOCKET_ERROR) {
	log_debug ("do_userpass_auth failed ec=%d\n", (int)WSAGetLastError ());
	free (pkt);
	return -1;
    }
    /* Parse reply [VER 1|STATUS 1] */
    if (recv (conn_fd, (char*)pkt, 2, 0) == SOCKET_ERROR) {
	log_debug ("do_userpass_auth failed ec=%d\n", (int)WSAGetLastError ());
	free (pkt);
	return -1;
    }
    *err = pkt[1];
    free (pkt);
    if (*err != ERR_SUCCESS)
	return -1;
    return 0;
}


/* Do the actual authentication. */
static int
do_auth (int conn_fd, int auth_method, keyserver_proxy_t ctx, int *err)
{
    switch (auth_method) {
    case PROXY_AUTH_NONE:
	*err = 0;
	return 0;

    case PROXY_AUTH_PLAIN:	
	return do_userpass_auth (conn_fd, ctx, err);

    default:
	return -1;
    }
    return 0;
}


/*
       +----+-----+-------+------+----------+----------+
        |VER | CMD |  RSV  | ATYP | DST.ADDR | DST.PORT |
        +----+-----+-------+------+----------+----------+
        | 1  |  1  | X'00' |  1   | Variable |    2     |
        +----+-----+-------+------+----------+----------+
    Send a connection reqeust for host @host and port @port.
*/
static int
send_connect_request (int conn_fd, const char *host, WORD port)
{
    struct hostent *hp;
    DWORD addr;
    BYTE pkt[10];

    addr = inet_addr (host);
    if (addr == INADDR_NONE) {
	hp = gethostbyname (host);
	if (hp != NULL)
	    memcpy ((char*)&addr, hp->h_addr_list, hp->h_length);
	else {
	    log_debug ("gethostbyname() failed ec=%d\n", (int)WSAGetLastError ());
	    return -1;
	}
    }
    
    pkt[0] = 0x05;
    pkt[1] = 0x01; /*CONNECT*/
    pkt[2] = 0x00;
    pkt[3] = ADDR_IPV4;
    pkt[4] = (BYTE)addr >> 24;
    pkt[5] = (BYTE)addr >> 16;
    pkt[6] = (BYTE)addr >>  8;
    pkt[7] = (BYTE)addr >>  0;
    pkt[8] = (BYTE)port >>  8;
    pkt[9] = (BYTE)port >>  0;
    if (send (conn_fd, (const char*)pkt, 10, 0) == SOCKET_ERROR) {
	log_debug ("send_connect_request failed ec=%d\n", (int)WSAGetLastError ());
	return -1;
    }

    return 0;
}


/*
        +----+-----+-------+------+----------+----------+
        |VER | REP |  RSV  | ATYP | BND.ADDR | BND.PORT |
        +----+-----+-------+------+----------+----------+
        | 1  |  1  | X'00' |  1   | Variable |    2     |
        +----+-----+-------+------+----------+----------+
    Parse the connect reply.
    Set @err with the specific error value.
*/
static int
recv_connect_reply (int conn_fd, int *err)
{
    BYTE pkt[4], addr[16], port[2];
    int ndomain;

    *err = 0;
    if (recv (conn_fd, (char*)pkt, DIM (pkt), 0) == SOCKET_ERROR) {
	log_debug ("recv_connect_reply: recv() failed ec=%d\n", 
		    (int)WSAGetLastError ());
	return -1;
    }
    if (pkt[0] != 0x05 || pkt[2] != 0x00) {
	log_debug ("recv_connect_reply: protocol violation.\n");
	return -1;
    }
    *err = pkt[1];
    if (pkt[1] != ERR_SUCCESS)
	return -1;
    switch (pkt[3]) {
    case ADDR_IPV4:
	if (recv (conn_fd, (char*)addr, 4, 0) == SOCKET_ERROR)
	    return -1;
	break;

    case ADDR_DOMAIN:
	recv (conn_fd, (char*)addr, 1, 0);
	ndomain = addr[0];
	/* Ignore the output, just empty the socket buffer. */
	while (ndomain-- > 0)
	    recv (conn_fd, (char*)addr, 1, 0);
	break;

    case ADDR_IPV6:
	if (recv (conn_fd, (char*)addr, DIM (addr), 0) == SOCKET_ERROR)
	    return -1;
	break;
    }
    if (recv (conn_fd, (char*)port, DIM (port), 0) == SOCKET_ERROR)
	return -1;
    return 0;
}


/* Perform all needed steps to initiate a SOCKSv5 proxy connection .*/
int
socks_handshake (keyserver_proxy_t ctx, int conn_fd,
		 const char *hostname, WORD port)
{
    int method, err;

    if (send_methods (conn_fd))
	return -1;
    method = recv_method (conn_fd);
    if (method == METH_ERR) {
	msg_box (NULL, _("No acceptable methods"), _("SOCKS Proxy"), MB_ERR);
	return -1;
    }
    if (do_auth (conn_fd, method, ctx, &err)) {
	if (!err)
	    msg_box (NULL, _("Authentication failed or unspported"), 
		     _("SOCKS Proxy"), MB_ERR);
	else
	    msg_box (NULL, socks_strerror (err), _("SOCKS Proxy"), MB_ERR);
	return -1;
    }
    if (send_connect_request (conn_fd, hostname, port)) {
	msg_box (NULL, _("Could not send SOCKS request"), 
		 _("SOCKS Proxy"), MB_ERR);
	return -1;
    }
    if (recv_connect_reply (conn_fd, &err)) {
	if (!err)
	    msg_box (NULL, _("Error while parsing SOCKS reply"), 
		     _("SOCKS Proxy"), MB_ERR);
	else
	    msg_box (NULL, socks_strerror (err), _("SOCKS Proxy"), MB_ERR);
	return -1;
    }
    return 0;
}
