/* wptSafeEditCtrl.cpp - Safe edit control for passwords
 *	Copyright (C) 2006 Timo Schulz, g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <richedit.h>
#include "resource.h"

#include "wptTypes.h"
#include "wptW32API.h"

/* New message to retrieve text. */
#define WM_SAFE_GETTEXT (WM_USER+11)

/* Hook context. */
struct hook_ctx_s {
    HHOOK keyb;	    /* handle for the keyboard. */
    HHOOK cbt;	    /* handle for computer based training (CBT) */
};
typedef struct hook_ctx_s *hook_ctx_t;

/* Subclassing context. */
struct subclass_ctx_s {
    WNDPROC new_wnd_fnc;    /* old window procedure. */
    WNDPROC old_wnd_fnc;    /* new window procedure. */
    HWND wnd;		    /* window handle that is subclassed. */
    void *opaque;
};
typedef struct subclass_ctx_s *subclass_ctx_t;


/* Dummy hook procedure to avoid channing. */
static LRESULT CALLBACK 
dummy_hook_proc (int code, WPARAM wparam, LPARAM lparam)
{
    return FALSE;
}


/* Set dummy hooks to prevent that the control is monitored
   by some external hook functions. */
static hook_ctx_t
install_dummy_hooks (void)
{
    hook_ctx_t hc;

    hc = (hook_ctx_t)calloc (1, sizeof *hc);
    if (!hc)
	return NULL;
    hc->cbt = SetWindowsHookEx (WH_CBT, dummy_hook_proc,
				 GetModuleHandle (NULL), 0);
    if (hc->cbt == NULL) {
	free (hc);
	return NULL;
    }
    hc->keyb = SetWindowsHookEx (WH_KEYBOARD, dummy_hook_proc,
				 GetModuleHandle (NULL), 0);
    if (hc->keyb == NULL) {
	free (hc);
	return NULL;
    }
    return hc;
}


/* Remove dummy hooks and free memory. */
static void
deinstall_dummy_hooks (hook_ctx_t hc)
{
    UnhookWindowsHookEx (hc->cbt);
    UnhookWindowsHookEx (hc->keyb);
    free (hc);
}


/* Filtered window procedure for the passphrase edit control. */
static LRESULT CALLBACK 
safe_edit_dlg_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static subclass_ctx_t ctx = NULL;

    /* XXX: filter all messages which were not send from 
            the current process. */

    if (ctx == NULL)
	ctx = (subclass_ctx_t)GetWindowLong (hwnd, GWL_USERDATA);

    switch (msg) {
    case WM_CUT:
    case WM_COPY:
    case WM_PASTE:
	/* do not allow to copy or paste the inserted text. */
	return TRUE;

    case WM_KEYUP:
    case WM_KEYDOWN:
	/* XXX: maybe in a future release, we might want to provide
	        our own buffer implementation. */

	/* ignore 'end' and 'home' to make text selection more difficult. */
	if ((int)wparam == VK_END || (int)wparam == VK_HOME)
	    return TRUE;
	break;

    case WM_GETTEXT:
	/* disallow normal gettext message. */
	return TRUE;

    case WM_SAFE_GETTEXT:
	{ /* transform WM_GETTEXT request into EM_GETLINE request. */
	    char *buffer = (char*)wparam;
	    int n;
	    buffer[0] = (char)lparam;
	    n = SendMessage (hwnd, EM_GETLINE, 0, wparam);
	    buffer[n] = 0;
	}
	break;

    case EM_GETLINE:
	break;

    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONUP:
    case WM_RBUTTONDOWN:
	/* prevent that the popup menu is displayed. */
	return TRUE;

    case EM_SETSEL:
    case EM_SCROLLCARET:
    case EM_FINDTEXT:
    case EM_FINDTEXTEX:
    case EM_GETSELTEXT:
    case EM_PASTESPECIAL:
    case EM_GETTEXTRANGE:
    case EM_STREAMIN:
    case EM_STREAMOUT:
	/* disallow message which could be used to copy the text. */
	return TRUE;
    }

    return CallWindowProc (ctx->old_wnd_fnc, hwnd, msg, wparam, lparam);
}


/* Sublclass the window in the dialog @dlg with the id @ctrlid.
   @new_wnd_fnc is the new window procedure. */   
static int
subclass_window (HWND dlg, int ctrlid, WNDPROC new_wnd_fnc,
		 subclass_ctx_t *r_ctx)
{
    subclass_ctx_t ctx;

    *r_ctx = NULL;
    ctx = (subclass_ctx_t)calloc (1, sizeof *ctx);
    if (!ctx)
	return -1;
    ctx->wnd = GetDlgItem (dlg, ctrlid);
    ctx->new_wnd_fnc = new_wnd_fnc;
    ctx->old_wnd_fnc = (WNDPROC)GetWindowLong (ctx->wnd, GWL_WNDPROC);
    if (!ctx->old_wnd_fnc) {
	free (ctx);
	return -1;
    }
    SetLastError (0);
    SetWindowLong (ctx->wnd, GWL_WNDPROC, (LONG)new_wnd_fnc);
    if (GetLastError () != 0) {
	free (ctx);
	return -1;
    }
    SetLastError (0);
    SetWindowLong (ctx->wnd, GWL_USERDATA, (LONG)ctx);
    if (GetLastError () != 0) {
	free (ctx);
	return -1;
    }
    if (r_ctx)
	*r_ctx = ctx;
    return 0;
}


/* Restore the old window procedure for the subclassed control
   and free the memory. */
static void
subclass_free_memory (HWND dlg, int ctrlid)
{
    subclass_ctx_t ctx;
    HWND wnd;

    wnd = GetDlgItem (dlg, ctrlid);
    ctx = (subclass_ctx_t)GetWindowLong (wnd, GWL_USERDATA);
    SetWindowLong (ctx->wnd, GWL_WNDPROC, (LONG)ctx->old_wnd_fnc);
    if (ctx != NULL)
	free (ctx);
}


/* Init subclassing to provide a safer edit control
   for the dialog @dlg control @ctlid. */
void
safe_edit_control_init (HWND dlg, int ctlid)
{
    subclass_ctx_t ctx;

    subclass_window (dlg, ctlid,  safe_edit_dlg_proc, &ctx);
    if (ctx != NULL)
	ctx->opaque = install_dummy_hooks ();
}


/* Reset subclassing for the edit control @ctlid. */
void
safe_edit_control_free (HWND dlg, int ctlid)
{
    subclass_ctx_t ctx;
    HWND wnd;

    wnd = GetDlgItem (dlg, ctlid);
    ctx = (subclass_ctx_t)GetWindowLong (wnd, GWL_USERDATA);
    if (ctx != NULL && ctx->opaque != NULL)
	deinstall_dummy_hooks ((hook_ctx_t)ctx->opaque);
    subclass_free_memory (dlg, ctlid);
}


UINT
SafeGetDlgItemText (HWND dlg, int id, char *buf, int buflen)
{
    memset (buf, 0, buflen);
    SendDlgItemMessage (dlg, id, WM_SAFE_GETTEXT, (WPARAM)buf, (LPARAM)buflen);
    return strlen (buf);
}


